from collections import defaultdict
import numpy as np
from future.utils import string_types
from .helpers import load_regions, load_sparse_matrix


def _chromosome_bins(regions):
    chr_bins = {}
    for r in regions:
        if chr_bins.get(r.chromosome) is None:
            chr_bins[r.chromosome] = [r.ix, r.ix + 1]
        else:
            chr_bins[r.chromosome][1] = r.ix + 1
    return chr_bins


def possible_contacts(regions, mappability):
    chromosomes = []
    current_chromosome = None
    for r in regions:
        if r.chromosome != current_chromosome:
            chromosomes.append(r.chromosome)
        current_chromosome = r.chromosome

    cb = _chromosome_bins(regions)
    chromosome_max_distance = defaultdict(int)
    max_distance = 0
    chromosome_subtractions = dict()
    for chromosome in chromosomes:
        start, stop = cb[chromosome]
        max_distance = max(max_distance, stop - start)
        chromosome_max_distance[chromosome] = max(chromosome_max_distance[chromosome], stop - start)
        chromosome_subtractions[chromosome] = np.zeros(stop - start, dtype='int32')

    chromosome_mappable = defaultdict(int)
    chromosome_unmappable = defaultdict(set)
    for i, mappable in enumerate(mappability):
        chromosome = regions[i].chromosome
        if not mappable:  # unmappable
            s = chromosome_subtractions[chromosome]
            o = cb[chromosome][0]
            ix = i - o
            # horizontal
            s[0: len(s) - ix] += 1
            # vertical
            for j in range(1, ix + 1):
                if ix - j not in chromosome_unmappable[chromosome]:
                    s[j] += 1
            chromosome_unmappable[chromosome].add(ix)
        else:
            chromosome_mappable[chromosome] += 1

    inter_total = 0
    intra_total = [0] * max_distance
    chromosome_intra_total = dict()
    for chromosome, d in chromosome_max_distance.items():
        chromosome_intra_total[chromosome] = [0] * d

    for i, chromosome in enumerate(chromosomes):
        start, stop = cb[chromosome]
        count = stop - start

        # intra-chromosomal
        s = chromosome_subtractions[chromosomes[i]]
        for distance in range(0, count):
            intra_total[distance] += count - distance - s[distance]
            chromosome_intra_total[chromosome][distance] += count - distance - s[distance]

        # inter-chromosomal
        for j in range(i + 1, len(chromosomes)):
            count_mappable = chromosome_mappable[chromosomes[i]]
            count2_mappable = chromosome_mappable[chromosomes[j]]
            inter_total += count_mappable * count2_mappable

    return intra_total, chromosome_intra_total, inter_total


def expected_values(regions, sparse_matrix, selected_chromosome=None):
    # get all the bins of the different chromosomes
    chromosome_bins = _chromosome_bins(regions)
    chromosome_dict = defaultdict(list)

    chromosome_max_distance = defaultdict(int)
    max_distance = 0
    for chromosome, (start, stop) in chromosome_bins.items():
        max_distance = max(max_distance, stop - start)
        chromosome_max_distance[chromosome] = max(chromosome_max_distance[chromosome], stop - start)

        for i in range(start, stop):
            chromosome_dict[i] = chromosome

    chromosome_intra_sums = dict()
    chromosome_intra_expected = dict()
    for chromosome, d in chromosome_max_distance.items():
        chromosome_intra_sums[chromosome] = [0.0] * d
        chromosome_intra_expected[chromosome] = [0.0] * d

    # get the sums of edges at any given distance
    marginals = [0.0] * len(regions)
    inter_sums = 0.0
    intra_sums = [0.0] * max_distance
    for source, sink, weight in sparse_matrix:
        if not np.isfinite(weight):
            continue

        source_chromosome = chromosome_dict[source]
        sink_chromosome = chromosome_dict[sink]

        marginals[source] += weight
        marginals[sink] += weight

        if sink_chromosome != source_chromosome:
            inter_sums += weight
        else:
            distance = sink - source
            intra_sums[distance] += weight
            chromosome_intra_sums[source_chromosome][distance] += weight

    mappability = [m > 0 for m in marginals]
    intra_total, chromosome_intra_total, inter_total = possible_contacts(regions, mappability)

    # expected values
    inter_expected = 0 if inter_total == 0 else inter_sums / inter_total

    intra_expected = [0.0] * max_distance
    bin_size = regions[0].end - regions[0].start
    distances = []
    for d in range(max_distance):
        distances.append(bin_size * d)

        # whole genome
        count = intra_total[d]
        if count > 0:
            intra_expected[d] = intra_sums[d] / count

    # chromosomes
    for chromosome in chromosome_intra_expected:
        for d in range(chromosome_max_distance[chromosome]):
            chromosome_count = chromosome_intra_total[chromosome][d]
            if chromosome_count > 0:
                chromosome_intra_expected[chromosome][d] = chromosome_intra_sums[chromosome][d] / chromosome_count

    if selected_chromosome is not None:
        return chromosome_intra_expected[selected_chromosome]

    return intra_expected, chromosome_intra_expected, inter_expected


def _oe_generator(sparse_matrix, ix_to_chromosome, intra_expected, inter_expected, log=False):
    for source, sink, weight in sparse_matrix:
        source_chromosome = ix_to_chromosome[source]
        sink_chromosome = ix_to_chromosome[sink]
        if source_chromosome == sink_chromosome:
            e = intra_expected[source_chromosome][sink - source]
        else:
            e = inter_expected

        try:
            oe = weight/e
        except ZeroDivisionError:
            oe = 1

        if log:
            yield source, sink, np.log2(oe)

        yield source, sink, oe


def observed_expected(regions, sparse_matrix):
    ix_converter = None
    if isinstance(regions, string_types):
        regions, ix_converter, _ = load_regions(regions)

    if isinstance(sparse_matrix, string_types):
        sparse_matrix = load_sparse_matrix(sparse_matrix, ix_converter=ix_converter)

    _, chromosome_intra_expected, inter_expected = expected_values(regions, sparse_matrix)

    ix_to_chromosome = {i: r.chromosome for i, r in enumerate(regions)}

    return regions, list(_oe_generator(sparse_matrix, ix_to_chromosome, chromosome_intra_expected, inter_expected))



import argparse
import textwrap
import sys


def chess_parser():
    usage = '''\
        chess <command> [options]

        Commands:
            sim             Calculate structural similarity
            oe              Transform a Hi-C matrix to an observed/expected matrix
            pairs           Make window pairs for chess genome scan
            background   Generate background region BED files
            extract         Extract specific regions that are significantly different
            crosscorrelate  Get structural clusters from the extracted submatrices

        Run chess <command> -h for help on a specific command.
        '''
    parser = argparse.ArgumentParser(
        description='CHESS: Compare Hi-C Experiments using Structural Similarity',
        usage=textwrap.dedent(usage)
    )

    from .version import __version__
    parser.add_argument('--version', action='version', version=f'CHESS {__version__}')

    parser.add_argument(
        '--verbose', '-v', dest='verbosity',
        action='count',
        default=0,
        help='Set verbosity level: Can be chained like '
             '"-vvv" to increase verbosity. Default is to show '
             'errors, warnings, and info messages (same as "-vv"). '
             '"-v" shows only errors and warnings, "-vvv" shows errors '
             'warnings, info, and debug messages in addition.')

    parser.add_argument(
        '-l', '--log-file', dest='log_file',
        help='Path to file in which to save log.'
    )

    parser.add_argument('command', nargs='?', help='Subcommand to run')

    return parser


def sim_parser():
    parser = MyParser(
            description='Compare structures between pairs of chromatin '
                        'contact matrices using the structural similarity '
                        'index. If --background-query or --background-regions '
                        'are specified, compute p-value and z-value '
                        'after obtaining a background distribution '
                        'of similarity values of the reference in the pair '
                        'to specfied background regions. '
                        'The input matrices are expected to be balanced, '
                        'e.g. by Knight-Ruiz matrix balancing. '
                        'The contacts are automatically converted internally '
                        'to observed / expected values. In your input files '
                        'are already observed / expected transformed, '
                        'set the --oe-input flag.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        'reference_contacts',
        type=str,
        help='Balanced contact matrix for the reference sample '
             'in one of the following formats: fanc .hic, '
             'juicer .hic@<resolution>,'
             'cooler .cool@<resolution> or .mcool@<resolution>, '
             'sparse format '
             '(each line: '
             '<row region index> <column region index> <matrix value>). '
             'If the file is in sparse format, the corresponding regions '
             ' BED file needs to be passed via --reference-regions.')

    parser.add_argument(
        '--reference-regions', dest="reference_regions",
        type=str,
        help='BED file (no header) with regions corresponding to '
             'the number of rows in the provided reference matrix.')

    parser.add_argument(
        'query_contacts',
        type=str,
        help='Balanced contact matrix for the query sample '
             'in one of the following formats: fanc .hic, '
             'juicer .hic@<resolution>,'
             'cooler .cool@<resolution> or .mcool@<resolution>, '
             'sparse format '
             '(each line: '
             '<row region index> <column region index> <matrix value>). '
             'If the file is in sparse format, the corresponding regions '
             ' BED file needs to be passed via --query-regions.')

    parser.add_argument(
        '--query-regions', dest="query_regions",
        type=str,
        help='BED file (no header) with regions corresponding to '
             'the number of rows in the provided query matrix.')

    parser.add_argument(
        'pairs',
        type=str,
        help='Region pairs to compare. '
             'Expected to be in 10 column BEDPE format as defined in the BEDTools docs: '
             'https://bedtools.readthedocs.io/en/latest/content/general-usage.html, with chrom1, start1, etc '
             'corresponding to reference and chrom2, start2, etc to query. The "score" column is not used, but should '
             'be present (can be "." or another placeholder value). All other columns are required.')

    parser.add_argument(
        'out',
        type=str,
        help='Path to outfile.')

    parser.add_argument(
        '--background-regions', dest='background_regions',
        type=str,
        help='BED file with regions to be used for background calculations. '
             'If provided, CHESS will generate Z-scores and P-values for '
             'similarities.')

    parser.add_argument(
        '--background-query', dest='background_query',
        default=False,
        action='store_true',
        help='Use every region of the same size as the '
             'reference from the query genome as background. '
             'Useful, for example, as background for inter-species '
             'comparisons.')

    parser.add_argument(
        '-p', dest='threads',
        type=int,
        default=1,
        help='Number of cores to use.')

    parser.add_argument(
        '--keep-unmappable-bins',
        action='store_true',
        default=False,
        help='Disable deletion of unmappable bins.')

    parser.add_argument(
        '--mappability-cutoff',
        type=float,
        default=0.1,
        help='Low pass threshold for fraction of unmappable bins. '
             'Matrices with a higher content of unmappable bins '
             'will not be considered. '
             'Unmappable bins will be deleted from matrices '
             'passing the filter.')

    parser.add_argument(
        '-r', '--relative-windowsize',
        type=float,
        default=1,
        help='Relative window size value '
             'for the win_size param in the ssim function. '
             'Fraction of the matrix size.')

    parser.add_argument(
        '-a', '--absolute-windowsize',
        type=int,
        default=None,
        help='Absolute window size value in bins '
             'for the win_size param in the ssim function. '
             'Overwrites -r.')

    parser.add_argument(
        '--oe-input',
        action='store_true',
        default=False,
        help='Use if input contacts are already observed / expected '
             'transformed.')

    parser.add_argument(
        '--limit-background',
        action='store_true',
        default=False,
        help='Restrict background computation to the syntenic / paired '
             'chromosome as indicated in the pairs file.')

    return parser


def oe_parser():
    parser = MyParser(
        description='Convert a sparse Hi-C matrix to '
                    'observed/expected format.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        'input_matrix',
        type=str,
        help='Balanced Hi-C matrix in sparse format. '
             '(each line: '
             '<row region index> <column region index> <matrix value>)')

    parser.add_argument(
        'regions',
        type=str,
        help='BED file (no header) with regions corresponding to '
             'the number of rows in the provided reference matrix.')

    parser.add_argument(
        'output_matrix',
        type=str,
        help='Obs/exp transformed matrix (same as input matrix format)')

    return parser


def pairs_parser():
    parser = MyParser(
        description='Make window pairs for CHESS genome scan. '
                    'Write all positions of a sliding window of specified '
                    'size with specified step in the specified genome to '
                    'the outfile which can be directly used to run CHESS sim.'
                    'Output will be in 10 column BEDPE format as defined in the BEDTools docs: '
                    'https://bedtools.readthedocs.io/en/latest/content/general-usage.html.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        'genome',
        type=str,
        help='UCSC genome identifier (as recognized by pybedtools), '
             'or path to tab-separated chrom sizes file with columns '
             '<chromosome name> <chromosome size>. '
             'Will use the path only if no UCSC entry with that name is found'
             ', or --file-input is specified.')

    parser.add_argument(
        'window',
        type=int,
        help='''Window size in base pairs''')

    parser.add_argument(
        'step',
        type=int,
        help='''Step size in base pairs''')

    parser.add_argument(
        'output',
        type=str,
        help='''Path to output file''')

    parser.add_argument(
        '--file-input',
        action='store_true',
        default=False,
        help='Will not check for UCSC entry of genome'
             'input with pybedtools if set'
        )

    parser.add_argument(
        '--chromosome',
        type=str,
        help='Produce window pairs only for the specified chromosome')

    return parser


def background_parser():
    parser = MyParser(
        description='Generate BED file with regions to be used'
                    'in CHESS background calculations.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        'genome_or_region',
        type=str,
        help='UCSC genome identifier (as recognized by pybedtools), '
             'OR path to tab-separated chrom sizes file with columns '
             '<chromosome name> <chromosome size> OR region identifier in '
             'the format <chromosome>:<start>-<end>. '
             'Will try options in the order listed.')

    parser.add_argument(
        'window',
        type=int,
        help='Window size in base pairs')

    parser.add_argument(
        'step',
        type=int,
        help='Step size in base pairs')

    parser.add_argument(
        'output',
        type=str,
        help='Path to output file')

    parser.add_argument(
        '--strand',
        type=str,
        help='[+/-] .Generate regions on this strand only. '
             'Default: both strands')

    return parser


def extract_parser():
    parser = MyParser(
        description='Extract the specific regions that are different '
                    'between the regions identified by CHESS.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        'pairs',
        type=str,
        help='Region pairs that have been identified to contain structural differences.'
             'Expected to be in 10 column BEDPE format as defined in the BEDTools docs: '
             'https://bedtools.readthedocs.io/en/latest/content/general-usage.html, with chrom1, start1, etc '
             'corresponding to reference and chrom2, start2, etc to query. The "score" column is not used, but should '
             'be present (can be "." or another placeholder value). All other columns are required.')

    parser.add_argument(
        'reference_contacts',
        type=str,
        help='Balanced contact matrix for the reference sample '
             'in one of the following formats: fanc .hic, '
             'juicer .hic@<resolution>,'
             'cooler .cool@<resolution> or .mcool@<resolution>, '
             'sparse format '
             '(each line: '
             '<row region index> <column region index> <matrix value>). '
             'If the file is in sparse format, the corresponding regions '
             ' BED file needs to be passed via --reference-regions.')

    parser.add_argument(
        '--reference-regions', dest="reference_regions",
        type=str,
        help='BED file (no header) with regions corresponding to '
             'the number of rows in the provided reference matrix.')

    parser.add_argument(
        'query_contacts',
        type=str,
        help='Balanced contact matrix for the query sample '
             'in one of the following formats: fanc .hic, '
             'juicer .hic@<resolution>,'
             'cooler .cool@<resolution> or .mcool@<resolution>, '
             'sparse format '
             '(each line: '
             '<row region index> <column region index> <matrix value>). '
             'If the file is in sparse format, the corresponding regions '
             ' BED file needs to be passed via --query-regions.')

    parser.add_argument(
        '--query-regions', dest="query_regions",
        type=str,
        help='BED file (no header) with regions corresponding to '
             'the number of rows in the provided query matrix.')

    parser.add_argument(
        'out',
        type=str,
        help='Path to output directory.')

    parser.add_argument(
        '--windowsize',
        type=int,
        default=3,
        help='Window size to average the bins according to their spatial '
             'closeness and their radiometric similarity, '
             'by default the windows size is the 3 x 3 bins. '
             'Larger values will average bins with larger differences.')

    parser.add_argument(
        '--sigma-spatial', dest="sigma_spatial",
        type=int,
        default=3,
        help='Gaussian function of the Euclidean distance between '
             'two bins and its standard deviation. '
             'Larger values will average bins with larger differences.')

    parser.add_argument(
        '--size-medianfilter', dest="size_medianfilter",
        type=int,
        default=9,
        help='Windows size used to scan and smooth the contained bins. '
             'Higher values will smooth larger figures, '
             'while smaller values will consider subtle signals (i.e. loops).')

    parser.add_argument(
        '--closing-square', dest="closing_square",
        type=int,
        default=8,
        help='Side length of the square used to remove noise, '
             'and fill structures. '
             'Larger values will enclose larger structures and remove '
             'punctuate or looping structures.')

    return parser


def crosscorrelate_parser():
    parser = MyParser(
        description='2D crosscorrelation of the specific substructures '
                    'extracted from the '
                    'regions with differences in their chromatin structures.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument(
        'extracted_file',
        type=str,
        help='Output from extract sub-command.')

    parser.add_argument(
        'pairs',
        type=str,
        help='Region pairs that have been identified to contain structural differences. '
             'Expected to be in 10 column BEDPE format as defined in the BEDTools docs: '
             'https://bedtools.readthedocs.io/en/latest/content/general-usage.html, with chrom1, start1, etc '
             'corresponding to reference and chrom2, start2, etc to query. The "score" column is not used, but should '
             'be present (can be "." or another placeholder value). All other columns are required.')

    parser.add_argument(
        'outdir',
        type=str,
        help='Path to output directory.')

    return parser


class MyParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)



import io
import gzip
import numpy as np
import os
import intervaltree
from future.utils import string_types
from collections import defaultdict
import logging
logger = logging.getLogger(__name__)


class GenomicRegion(object):
    """
    Class representing a genomic region.
    .. attribute:: chromosome
        Name of the chromosome this region is located on
    .. attribute:: start
        Start position of the region in base pairs
    .. attribute:: end
        End position of the region in base pairs
    .. attribute:: strand
        Strand this region is on (+1, -1)
    .. attribute:: ix
        Index of the region in the context of all genomic
        regions.
    """

    def __init__(self, start, end, chromosome=None, ix=None, strand=None):
        """
        Initialize this object.
        :param start: Start position of the region in base pairs
        :param end: End position of the region in base pairs
        :param chromosome: Name of the chromosome this region is located on
        :param ix: Index of the region in the context of all genomic
                   regions.
        """
        self.start = start
        self.end = end
        self.chromosome = chromosome
        self.ix = ix
        self.strand = strand

    @classmethod
    def from_string(cls, region_string):
        """
        Convert a string into a :class:`~GenomicRegion`.
        This is a very useful convenience function to quickly
        define a :class:`~GenomicRegion` object from a descriptor
        string.
        :param region_string: A string of the form
                              <chromosome>[:<start>-<end>[:<strand>]]
                              (with square brackets indicating optional
                              parts of the string). If any optional
                              part of the string is omitted, intuitive
                              defaults will be chosen.
        :return: :class:`~GenomicRegion`
        """
        chromosome = None
        start = None
        end = None

        # strip whitespace
        no_space_region_string = "".join(region_string.split())
        fields = no_space_region_string.split(':')

        if len(fields) > 3:
            raise ValueError(
                ("Genomic range string must be of the form"
                 " <chromosome>[:<start>-<end>:[<strand>]]"))

        # there is chromosome information
        if len(fields) > 0:
            chromosome = fields[0]

        # there is range information
        if len(fields) > 1 and fields[1] != '':
            start_end_bp = fields[1].split('-')
            if len(start_end_bp) > 0:
                try:
                    start = int(start_end_bp[0])
                except ValueError:
                    raise ValueError("Start of genomic range must be integer")

            if len(start_end_bp) > 1:
                try:
                    end = int(start_end_bp[1])
                except ValueError:
                    raise ValueError("End of genomic range must be integer")

                if not end > start:
                    raise ValueError(
                        "The end coordinate must be bigger than the start.")

        return cls(start=start, end=end, chromosome=chromosome)

    def overlaps(self, region):
        """
        Check if this region overlaps with the specified region.
        :param region: :class:`~GenomicRegion` object or string
        """
        if isinstance(region, str):
            region = GenomicRegion.from_string(region)

        if region.chromosome != self.chromosome:
            return False

        if region.start <= self.end and region.end >= self.start:
            return True
        return False

    def contains(self, region):
        """
        Check if the specified region is completely contained in this region.
        :param region: :class:`~GenomicRegion` object or string
        """
        if isinstance(region, str):
            region = GenomicRegion.from_string(region)

        if region.chromosome != self.chromosome:
            return False

        if region.start >= self.start and region.end <= self.end:
            return True
        return False

    def _equals(self, region):
        if region.chromosome != self.chromosome:
            return False
        if region.start != self.start:
            return False
        if region.end != self.end:
            return False
        return True

    def __eq__(self, other):
        return self._equals(other)

    def __ne__(self, other):
        return not self._equals(other)

    def __str__(self):
        return '{}:{}-{}:{}'.format(
            self.chromosome, self.start, self.end, self.strand)

    def copy(self):
        """
        Make a copy of this GenomicRegion object.
        """
        return GenomicRegion(
            chromosome=self.chromosome, start=self.start, end=self.end)


def is_gzipped(file_name):
    if file_name.endswith('.gz') or file_name.endswith('.gzip'):
        return True
    return False


def _open(file_name, mode='r'):
    if is_gzipped(file_name):
        return io.TextIOWrapper(gzip.open(file_name, mode=mode))
    return open(file_name, mode=mode)


def load_regions(file_name, sep=None, ignore_ix=False):
    """
    Load regions from regions bed file.
    :param file_name: Path to regions bed file.
    :param sep: Delimiter in regions bed, defaults to None (splits by tab)
    :returns: Tuple:
                (
                    List of :class:`~GenomicRegion` objects,
                    ix_converter dict: {region_id: position_in_list},
                    ix2region dict: {region_id: [chromosome, start, end]}
                (
    """
    regions = []
    ix2reg = {}
    ix_converter = None
    with _open(file_name, 'r') as f:
        for i, line in enumerate(f):
            line = line.rstrip()
            fields = line.split(sep)
            if len(fields) > 2:
                chromosome = fields[0]
                start = int(fields[1])
                end = int(fields[2])
                ix = i
                ix2reg[ix] = [chromosome, int(fields[1]), int(fields[2])]
                if not ignore_ix and len(fields) > 3 and fields[3] != '.':  # HicPro
                    if ix_converter is None:
                        ix_converter = dict()
                    if fields[3] in ix_converter:
                        raise ValueError(
                            "name column in region BED must "
                            "only contain unique values! ({})".format(
                                fields[3]))
                    ix_converter[fields[3]] = ix
                regions.append(
                    GenomicRegion(
                        chromosome=chromosome, start=start, end=end, ix=ix))

    if ix_converter is not None:
        ix_converter_rev = {str(v): int(k) for k, v in ix_converter.items()}
        ix2reg = {ix_converter_rev[str(k)]: v for k, v in ix2reg.items()}
    return regions, ix_converter, ix2reg


def region_interval_trees(regions):
    chromosome_intervals = defaultdict(list)
    for region in regions:
        interval = intervaltree.Interval(region.start - 1, region.end, data=region)
        chromosome_intervals[region.chromosome].append(interval)

    region_trees = dict()
    for chromosome, intervals in chromosome_intervals.items():
        region_trees[chromosome] = intervaltree.IntervalTree(intervals)

    return region_trees


def sub_regions(regions, region):
    """
    Get regions from a list the overlap with another region.
    :param regions: List of :class:`~GenomicRegion`
    :param region: :class:`~GenomicRegion` used for overlap calculation
    """
    if isinstance(region, string_types):
        region = GenomicRegion.from_string(region)

    # interval tree?
    try:
        rt = regions[region.chromosome]

        ixs = []
        sr = []
        for interval in rt[region.start - 1: region.end]:
            hit_region = interval.data
            ixs.append(hit_region.ix)
            sr.append(hit_region)
        start_ix = min(ixs)
        end_ix = max(ixs)

    # region list
    except TypeError:
        sr = []
        start_ix = None
        end_ix = None
        for i, r in enumerate(regions):
            if (r.chromosome == region.chromosome
                    and r.start <= region.end and r.end >= region.start):
                if start_ix is None:
                    start_ix = i
                end_ix = i
                sr.append(r.copy())
            else:
                if end_ix is not None:
                    break
    except IndexError:
        sr = []
        start_ix = None
        end_ix = None

    if len(sr) == 0:
        raise ValueError("Region not found in dataset! {}:{}-{}".format(
            region.chromosome, region.start, region.end))

    return sr, start_ix, end_ix


def sub_matrix_regions(hic_matrix, regions, region):
    """
    Get a square sub Hi-C matrix that overlaps a given region.
    :param hic_matrix: A square numpy array
    :param regions: List of :class:`~GenomicRegion`
    :param region: :class:`~GenomicRegion` used for overlap calculation
    """
    sr, start_ix, end_ix = sub_regions(regions, region)

    if start_ix is None:
        return np.empty((0, 0)), sr

    return np.copy(hic_matrix[start_ix:end_ix+1, start_ix:end_ix+1]), sr


def sub_matrix_from_edges_dict(edges, regions, region, default_weight=0.0):
    """
    Get a square sub Hi-C matrix that overlaps a given region.
    :param edges: A dict of the form { source_ix1: { sink_ix1: weight1, sink_ix2: weight2, ...}, ...}
    :param regions: List of :class:`~GenomicRegion` or dict of :class:`~intervaltree.IntervalTree`
    :param region: :class:`~GenomicRegion` used for overlap calculation
    :param default_weight: Default value with which to fill matrix
    """
    sr, start_ix, end_ix = sub_regions(regions, region)

    size = end_ix - start_ix + 1
    m = np.full((size, size), default_weight, dtype='float64')
    for i in range(start_ix, end_ix + 1):
        for j in range(i, end_ix + 1):
            try:
                m[i - start_ix, j - start_ix] = edges[i][j]
                m[j - start_ix, i - start_ix] = edges[i][j]
            except KeyError:
                pass

    return m, sr


def edges_from_sparse_matrix(file_name, ix_converter=None, sep="\t"):
    nan_values = 0
    total_values = 0

    with _open(file_name, 'r') as f:
        for line in f:
            if line.startswith("#"):
                continue
            total_values += 1

            line = line.rstrip()
            fields = line.split(sep)
            if ix_converter is None:
                source, sink, weight = int(fields[0]), int(fields[1]), float(fields[2])
            else:
                source = ix_converter[fields[0]]
                sink = ix_converter[fields[1]]
                weight = float(fields[2])

            if not np.isfinite(weight):
                nan_values += 1
                continue

            if source <= sink:
                yield source, sink, weight
            else:
                yield sink, source, weight

    if nan_values/total_values > 0.05:
        logger.warning("Could not import more than 5% of matrix entries ({}/{}), "
                       "as they were not finite".format(nan_values, total_values))


def edges_dict_from_sparse(edges):
    """Constructs edges_dict from edges parsed
    from sparse format data.

    Args:
        edges (iterator): iterator over (sink, source, weight) tuples

    Returns:
        dict: Nexted dictionary that allows
              queries like: dict[source][sink] = weight
    """
    edges_d = defaultdict(dict)
    for source, sink, weight in edges:
        edges_d[source][sink] = weight
    return edges_d


def edges_dict_from_fanc(hic):
    """Constructs edges_dict from edges
    present in a fanc hic object.

    Args:
        hic (fanc.Hic): Object created by fanc.load()

    Returns:
        dict: Nexted dictionary that allows
              queries like: dict[source][sink] = weight
    """
    edges_dict = defaultdict(dict)
    for e in hic.edges(lazy=True):
        edges_dict[e.source][e.sink] = e.weight
    return edges_dict


def oe_edges_dict_from_fanc(hic):
    """Constructs edges_dict from edges
    present in a fanc hic object. Weights are
    transformed to observed / expected values.

    Args:
        hic (fanc.Hic): Object created by fanc.load()

    Returns:
        dict: Nexted dictionary that allows
              queries like: dict[source][sink] = weight
    """
    edges_dict = defaultdict(dict)
    for e in hic.edges(oe=True, lazy=True):
        edges_dict[e.source][e.sink] = e.weight
    return edges_dict


def load_sparse_matrix(file_name, ix_converter=None, sep="\t"):
    return list(edges_from_sparse_matrix(file_name, ix_converter, sep))


def load_matrix(file_name, size=None, sep=None, ix_converter=None, is_oe=False):
    """Load Numpy matrix from Hi-C matrix file in sparse format.

    :param file_name: Path to Hi-C matrix file in sparse format.
    :param size: Size of the matrix (number of bins / edge length),
                 defaults to None
    :param sep: Delimiter in matrix file, defaults to None (splits by tab)
    :param ix_converter: ix_converter dict corresponding to the matrix file,
                         as produced by :func:`load_regions`, defaults to None
    :returns: Numpy matrix.
    """

    if size is None:
        raise ValueError("Must provide matrix size!")

    if is_oe:
        m = np.ones((size, size))
    else:
        m = np.zeros((size, size))

    for source, sink, weight in edges_from_sparse_matrix(
                                file_name, ix_converter=ix_converter, sep=sep):
        m[source, sink] = weight
        m[sink, source] = weight

    return m


def load_pairs(file_name, sep=None):
    """Load region pairs from bedpe file.

    Read input bedpe file and convert to dict mapping the comparison ID to the
    reference to query regions.
    :param file_name: Path to input bedpe file. Expected columns:
        chromosome_reference, start_reference, end_reference,
        chromosome_query, start_query, end_query,
        comparison_id, dummy column (not used), strand_reference, strand_query
    :param sep: Delimiter in bedpe file, defaults to None (splits by tab)
    :returns: Dict of form {ID: (reference_region, query_region)}, with
              *region being :class:`GenomicRegion` objects.
    """
    pairs = dict()
    for ID, cref, cqry in load_pairs_iter(file_name, sep=sep):
        pairs[ID] = (cref, cqry)
    return pairs


def load_pairs_iter(file_name, sep=None):
    """Load region pairs from bedpe file.

    Read input bedpe file and convert to dict mapping the comparison ID to the
    reference to query regions.
    :param file_name: Path to input bedpe file. Expected columns:
        chromosome_reference, start_reference, end_reference,
        chromosome_query, start_query, end_query,
        comparison_id, dummy column (not used), strand_reference, strand_query
    :param sep: Delimiter in bedpe file, defaults to None (splits by tab)
    :returns: Dict of form {ID: (reference_region, query_region)}, with
              *region being :class:`GenomicRegion` objects.
    """
    pairs = {}
    with _open(file_name, 'r') as f:
        for line in f:
            if line.startswith('#'):
                continue
            sline = line.rstrip()
            fields = sline.split(sep)
            if len(fields) != 10:
                raise ValueError((
                    "10 columns expected but {} found"
                    " in bedpe input"
                    ).format(len(fields)))
            c1, s1, e1, c2, s2, e2, ID, _, str1, str2 = line.split(sep)
            if ID in pairs:
                raise ValueError((
                    "Pair ID {} not unique in bedpe input!"))
            cref = GenomicRegion(
                chromosome=str(c1), start=int(s1) + 1,
                end=int(e1), strand=str(str1))
            cqry = GenomicRegion(
                chromosome=str(c2), start=int(s2) + 1,
                end=int(e2), strand=str(str2))
            yield ID, cref, cqry


def load_pairs_for_chroms(file_name, chromosomes, sep=None):
    """Load region pairs from bedpe file.

    Read input bedpe file and convert to dict mapping the comparison ID to the
    reference to query regions. Return only pairs which are on the specified
    chromosomes.

    :param file_name: Path to input bedpe file. Expected columns:
        chromosome_reference, start_reference, end_reference,
        chromosome_query, start_query, end_query,
        comparison_id, dummy column (not used), strand_reference, strand_query
    :param sep: Delimiter in bedpe file, defaults to None (splits by tab)
    :returns: Dict of form {ID: (reference_region, query_region)}, with
              *region being :class:`GenomicRegion` objects.
    """
    pairs = set()
    res = []
    dropped = 0
    with _open(file_name, 'r') as f:
        for line in f:
            if line.startswith('#'):
                continue
            sline = line.rstrip()
            fields = sline.split(sep)
            if len(fields) != 10:
                raise ValueError((
                    "10 columns expected but {} found"
                    " in bedpe input"
                    ).format(len(fields)))
            c1, s1, e1, c2, s2, e2, ID, _, str1, str2 = line.split(sep)
            if ID in pairs:
                raise ValueError((
                    "Pair ID {} not unique in bedpe input!"))
            pairs.add(ID)
            if any(c not in chromosomes for c in (c1, c2)):
                dropped += 1
            else:
                cref = GenomicRegion(
                    chromosome=str(c1), start=int(s1) + 1,
                    end=int(e1), strand=str(str1))
                cqry = GenomicRegion(
                    chromosome=str(c2), start=int(s2) + 1,
                    end=int(e2), strand=str(str2))
                res.append((ID, cref, cqry))
    return res, dropped


def chunks(l, p):
    n = int(np.ceil(len(l)/p))
    """Yield successive len(l)/p-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def read_chromosome_sizes(file_name):
    chrom_sizes = {}
    with open(os.path.expanduser(file_name), 'r') as chrom_sizes_file:
        for line in chrom_sizes_file:
            line = line.rstrip()
            if line != '':
                chromosome, chromosome_length = line.split("\t")
                chrom_sizes[chromosome] = int(chromosome_length)
    return chrom_sizes


def load_contacts(matrix_file, regions_file=None):
    import fanc
    from chess.oe import observed_expected
    try:
        # try loading via fanc
        reference_loaded = fanc.load(matrix_file)
        edges = oe_edges_dict_from_fanc(reference_loaded)
        regions = reference_loaded.regions
        region_trees = region_interval_trees(
            regions)
    except ValueError as initial_error:
        print(initial_error)
        try:
            assert regions_file is not None, (
                "Regions file needs to be"
                "specified for sparse input.")
            regions, _ix_converter, _ = load_regions(
                regions_file)
            region_trees = region_interval_trees(
                regions)
            _, reference_oe = observed_expected(
                regions_file, matrix_file)
            edges = edges_dict_from_sparse(reference_oe)
        except AssertionError as error:
            raise ValueError(error)
    return edges, region_trees, regions


def load_oe_contacts(matrix_file, regions_file=None):
    import fanc
    try:
        # try loading via fanc
        reference_loaded = fanc.load(matrix_file)
        edges = edges_dict_from_fanc(reference_loaded)
        regions = reference_loaded.regions
        region_trees = region_interval_trees(
            regions)
    except ValueError:
        try:
            assert regions_file is not None
            regions, ix_converter, _ = load_regions(
                regions_file)
            region_trees = region_interval_trees(
                regions)
            edges = edges_dict_from_sparse(
                edges_from_sparse_matrix(
                    matrix_file,
                    ix_converter))
        except AssertionError:
            raise ValueError
    return edges, region_trees, regions



#!/usr/bin/env python

from os import path
import logging
import numpy as np
from scipy import ndimage as ndi
from skimage import restoration
from skimage.morphology import square, closing
import skimage.filters as filters
from skimage.measure import label, regionprops
from numpy import inf
from scipy.ndimage import zoom
import warnings
from future.utils import string_types
from .helpers import (
    load_regions, load_sparse_matrix, sub_matrix_from_edges_dict
    )
warnings.filterwarnings("ignore")
logger = logging.getLogger(__name__)


def get_sparse_matrix(regions, sparse_matrix):
    '''Get all the sparse HiC matrix from the bed file'''
    ix_converter = None
    if isinstance(regions, string_types):
        regions, ix_converter, _ = load_regions(regions)

    if isinstance(sparse_matrix, string_types):
        sparse_matrix = load_sparse_matrix(
            sparse_matrix, ix_converter=ix_converter)

    return regions, sparse_matrix


def clipped_zoom(img, zoom_factor, **kwargs):
    h, w = img.shape[:2]
    zoom_tuple = (zoom_factor,) * 2 + (1,) * (img.ndim - 2)
    if zoom_factor < 1:
        zh = int(np.round(h * zoom_factor))
        zw = int(np.round(w * zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2
        out = np.zeros_like(img)
        out[top:top+zh, left:left+zw] = zoom(img, zoom_tuple, **kwargs)
    elif zoom_factor > 1:
        zh = int(np.round(h / zoom_factor))
        zw = int(np.round(w / zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2
        out = zoom(img[top:top+zh, left:left+zw], zoom_tuple, **kwargs)
        trim_top = ((out.shape[0] - h) // 2)
        trim_left = ((out.shape[1] - w) // 2)
        out = out[trim_top:trim_top+h, trim_left:trim_left+w]
    else:
        out = img
    return out


def get_info_feature(labels, submatrix, outfile, position, area, reg):
    for region in regionprops(labels):
        minr, minc, maxr, maxc = region.bbox
        if region.area > area:
            bx = (minc, maxc, maxc, minc, minc)
            by = (minr, minr, maxr, maxr, minr)
            y_min, y_max = np.min(by), np.max(by)
            x_min, x_max = np.min(bx), np.max(bx)
            submat = submatrix[y_min:y_max, x_min:x_max]
            flat_mat = list(submat.flatten())
            flat_mat.insert(0, reg)
            flat_mat.insert(1, position)
            flat_mat.insert(2, x_min)
            flat_mat.insert(3, x_max)
            flat_mat.insert(4, y_min)
            flat_mat.insert(5, y_max)
            outfile.write(','.join(map(str, flat_mat)) + '\n')
            position += 1
    return position


def extract_structures(
    reference_edges,
    reference_regions,
    query_edges,
    query_regions,
    pairs,
    output,
    windowsize,
    sigma_spatial,
    size_medianfilter,
    closing_square
):
    pos_query = 0
    pos_reference = 0

    w1 = open(path.join(output, 'gained_features.tsv'), '+w')
    w2 = open(path.join(output, 'lost_features.tsv'), '+w')

    for pair_ix, reference_region, query_region in pairs:
        reference, ref_rs = sub_matrix_from_edges_dict(
            reference_edges,
            reference_regions,
            reference_region,
            default_weight=0.)
        query, qry_rs = sub_matrix_from_edges_dict(
            query_edges,
            query_regions,
            query_region,
            default_weight=0.)
        area = int((5 * np.shape(query)[0]) / 100)
        size = np.shape(query)[0]
        or_matrix = np.log(np.divide(query, reference))
        where_are_NaNs = np.isnan(or_matrix)
        or_matrix[where_are_NaNs] = 0.
        or_matrix[or_matrix == -inf] = 0.
        or_matrix[or_matrix == inf] = 0.
        std = np.std(or_matrix)
        # thres1, thres2 = mean + 0.5 * std, mean - 0.5 * std
        # positive = np.where(or_matrix > thres1, or_matrix, 0.)
        # negative = np.where(or_matrix < thres2, or_matrix, 0.)
        positive = np.where(or_matrix > (0.5 * std), or_matrix, 0.)
        negative = np.abs(np.where(or_matrix < -(0.5 * std), or_matrix, 0.))

        # denoise
        denoise_positive = restoration.denoise_bilateral(
            positive,
            sigma_color=np.mean(positive),
            win_size=windowsize,
            sigma_spatial=sigma_spatial,
            bins=size,
            multichannel=False)
        denoise_negative = restoration.denoise_bilateral(
            negative,
            sigma_color=np.mean(negative),
            win_size=windowsize,
            sigma_spatial=sigma_spatial,
            bins=size,
            multichannel=False)
        # smooth
        filter_positive = ndi.median_filter(
            denoise_positive, size_medianfilter)
        filter_negative = ndi.median_filter(
            denoise_negative, size_medianfilter)

        # binarise
        if np.all(filter_positive == 0.):
            threshold_pos = positive
        else:
            filter1 = filters.threshold_otsu(filter_positive, nbins=size)
            threshold_pos = filter_positive > filter1

        if np.all(filter_negative == 0.):
            threshold_neg = negative
        else:
            filter2 = filters.threshold_otsu(filter_negative, nbins=size)
            threshold_neg = filter_negative > filter2

        # zoom and rotate 45 degrees
        zm1 = clipped_zoom(threshold_pos, 0.7)
        rot1 = ndi.rotate(zm1, 45, reshape=False)
        zm2 = clipped_zoom(threshold_neg, zoom_factor=0.7)
        rot2 = ndi.rotate(zm2, 45, reshape=False)

        # Close morphology
        img1 = closing(rot1, square(closing_square))
        label_x1 = label(img1)
        img2 = closing(rot2, square(closing_square))
        label_x2 = label(img2)

        zm1_reference = clipped_zoom(reference, 0.7)
        rot_reference = ndi.rotate(zm1_reference, 45, reshape=False)
        zm1_query = clipped_zoom(query, 0.7)
        rot_query = ndi.rotate(zm1_query, 45, reshape=False)

        # get output (file with label and submatrices)
        pos_query = get_info_feature(
            label_x1, rot_query, w1, pos_query, area, pair_ix)
        pos_reference = get_info_feature(
            label_x2, rot_reference, w2, pos_reference, area, pair_ix)


def rotate_feature(x, y, window_size, zoom_factor=1.0/0.7):
    """Rotate a single x,y coordinate within a region by -45 degrees.
    This allows conversion of coordinates from chess extract output to genomic bins.

    Parameters
    ----------
    x : int
        The x coordinate of a single point within the window, as found in the
        output of chess extract
    y : int
        The y coordinate of a single point within the window, as found in the
        output of chess extract.
    window_size : int
        The size of the window region.

    Returns
    -------
    x_r, y_r : tuple of int
        The rotated coordinates.
    
    Examples
    --------
    >>> rotate_feature(50, 50, 100) # stays in center
    (50, 50)
    >>> rotate_feature(20, 30, 60, zoom_factor=1) # vertical center -> diagonal
    (23, 23)
    """
    # NOTE: There will be some rounding errors when rotating the system. Not sure
    # what is the best way to handle them, for now I just round to the closest int.
    
    # Convert degree to radians
    theta = np.radians(-45)
    # Compute rotation matrix
    rot = np.array(
        [[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]]
    )
    # Get rotation center
    center = window_size / 2
    # Substract rotation center and rotate
    x_r, y_r = zoom_factor * rot @ np.array([x - center, y - center])
    # Add back rotation center
    x_r += center
    y_r += center
    return round(x_r), round(y_r)


def get_feature_coords(coords, win_size, win_bp_start, res):
    """Get basepair coordinates of a feature by rotating it and converting
    bins to coordinates

    Parameters
    ----------
    coords : tuple of int
        The coordinate of the feature within the window, as found in the
        output of chess extract: (xmin, xmax, ymin, ymax)
    win_size : int
        The size of the window region in bins.
    win_bp_start : int
        The start coordinate of the window in base pairs.
    res : int
        The resolution of the Hi-C matrix, i.e. the number of basepairs
        per bin.

    Returns
    -------
    x_coords, y_coords : tuple of lists of int
        The genomic coordinates of the feature on both axes, in
        the format: ([x_start, x_end], [y_start, y_end])
    
    Examples
    --------
    >>> chess_coords = (103, 108, 85, 90)
    >>> get_feature_coords(chess_coords, 201, 98000001, res=50000)
    ([103500001, 103650001], [102550001, 102750001])
    """
    # Rotate input coordinates by 45 degrees to get genomic bins
    xmin, xmax, ymin, ymax = coords
    xmin_r, ymin_r = rotate_feature(ymin, ymin, win_size)
    xmin_r, ymax_r = rotate_feature(xmin, ymax, win_size)
    xmax_r, ymin_r = rotate_feature(xmin, ymin, win_size)


    # Generate new genomic coordinates
    x_coords = [xmin_r * res + win_bp_start, xmax_r * res + win_bp_start]
    y_coords = [ymin_r * res + win_bp_start, ymax_r * res + win_bp_start]

    return x_coords, y_coords



#!/usr/bin/env python

import matplotlib
matplotlib.use('pdf')

import logging
from os import path
from collections import defaultdict
import numpy as np
import matplotlib.pyplot as plt
from sklearn.preprocessing import scale
from sklearn.cluster import KMeans
from scipy.signal.signaltools import correlate2d as c2d
import pandas as pd
import itertools
from tqdm import tqdm
from kneed import KneeLocator

"""Compute 2D cross-correlation and get clusters of structures"""


def correlate2d(file, output_folder, pairs):
    logger = logging.getLogger('')
    # load all submatrices and correlate them
    all_arrays = defaultdict(list)
    information_regions = defaultdict(list)
    pairs_dict = defaultdict(str)
    for pair_ix, reference_region, query_region in pairs:
        pairs_dict[int(pair_ix)] = str(reference_region)+'-'+str(query_region)

    with open(file, 'r') as r:
        for line in r:
            region, position, x_min, x_max, y_min, y_max = line.split(',')[:6]
            pair_id = pairs_dict[int(region)]
            line_float = [float(x) for x in line.split(',')[6:]]
            height, width = int(y_max) - int(y_min), int(x_max) - int(x_min)
            mat = np.asanyarray(line_float).reshape(int(height), int(width))
            all_arrays[int(position)].append(mat)
            information_regions[int(position)].append((pair_id, int(position)))

    logger.info(
        '[MAIN]: All submatrices loaded, starting 2D cross-correlation')
    tag = file.split('/')[-1].split('_')[0]
    correlation_dataframe = pd.DataFrame(
        index=range(len(all_arrays.keys())),
        columns=range(len(all_arrays.keys())))

    for a, b in tqdm(itertools.combinations(range(len(all_arrays.keys())), 2)):
        c11 = c2d(all_arrays[a][0], all_arrays[b][0], mode='same')
        c112 = c2d(all_arrays[b][0], all_arrays[a][0], mode='same')
        transp1 = c2d(
            np.fliplr(all_arrays[a][0]), all_arrays[b][0], mode='same')
        transp2 = c2d(
            np.fliplr(all_arrays[b][0]), all_arrays[a][0], mode='same')
        best = max(c11.max(), c112.max(), transp1.max(), transp2.max())
        correlation_dataframe.loc[a, b] = best
        correlation_dataframe.loc[b, a] = best
    logger.info('[MAIN]: 2D cross-correlation done')

    correlation_dataframe = correlation_dataframe.replace(np.Inf, np.nan)
    correlation_dataframe = correlation_dataframe.replace(-np.Inf, np.nan)
    scaled = scale(correlation_dataframe.fillna(0.))

    # save dataframe ##
    correlation_dataframe.to_csv(
        path.join(output_folder, 'correlation_dataframe_%s.csv' % (tag)))

    Sum_of_squared_distances = []
    K = range(1, 15)
    for k in K:
        km = KMeans(n_clusters=k)
        km = km.fit(scaled)
        Sum_of_squared_distances.append(km.inertia_)

    kn = KneeLocator(
        range(1, len(Sum_of_squared_distances)+1),
        Sum_of_squared_distances,
        curve='convex',
        direction='decreasing')
    optimal_number_clusters = kn.knee

    plt.figure(figsize=(4, 3))
    plt.xlabel('number of clusters k')
    plt.ylabel('Sum of squared distances')
    plt.plot(
        range(1, len(Sum_of_squared_distances)+1),
        Sum_of_squared_distances,
        'bx-')
    plt.vlines(kn.knee, plt.ylim()[0], plt.ylim()[1], linestyles='dashed')
    plt.savefig(path.join(output_folder, 'Elbow_index_%s.pdf' % tag))

    # save positions and belonging cluster
    logger.info('[MAIN]: Classification of the features')
    kmeans = KMeans(
        n_clusters=optimal_number_clusters,
        random_state=0,
        precompute_distances=True).fit(scaled)
    w = open(path.join(
        output_folder,
        'subregions_%s_clusters_%s.tsv' % (str(optimal_number_clusters), tag)),
        'a+')
    for n, i in enumerate(list(kmeans.labels_)):
        t1, t2 = information_regions[n][0]
        w.write('{}\t{}\t{}\n'.format('Cluster ' + str(i), t1, t2))
    w.close()



#!/usr/bin/env python

import logging
import numpy as np

from skimage.transform import resize
from skimage.metrics import structural_similarity
from numpy.lib.stride_tricks import as_strided
from .helpers import sub_matrix_from_edges_dict
import uuid

logger = logging.getLogger(__name__)


def SN(M1, M2, size=7):
    """Compute average of local signal to noise ratios in M1 - M2.

    Slide a window of dimensions size^2 over M1 - M2, compute
    abs(mean(window)) / var(window) in each and take mean of all windows.
    :param M1: Numpy Matrix.
    :param M2: Numpy Matrix.
    :param size: edge length of sliding window (must be odd), defaults to 7
    :returns: float, the average of the local signal to noise ratios.
    """
    M = M1 - M2
    uM = filter_uniform_filter(M, size=size)
    vM = filter_uniform_filter(M, size=size, func=np.nanvar)
    SN = np.nanmean(np.abs(uM) / vM)

    return float(SN)


def filter_uniform_filter(M, size=7, no_data_val=None, func=np.nanmean):
    """Apply func to sliding window on input matrix.

    Adaption from:
    www.programcreek.com/python/example/93943/scipy.ndimage.uniform_filter
    This is equivalent to scipy.ndimage.uniform_filter, but can handle nan's,
    and can use numpy nanmean/median/max/min functions.
    :param M: input data, numpy matrix.
    :param size: edge length of sliding window (must be odd), defaults to 7
    :type size: number, optional
    :param no_data_val: value in matrix that is treated as no data value,
                        defaults to None
    :param func: [description], defaults to np.nanmean
    :returns: func of the matrix A filtered by a uniform kernel of size=size
    """

    assert size % 2 == 1, 'Please supply an odd size'
    assert M.dtype == np.dtype('float64'), 'Input must have dtype(float64)'
    rows, cols = M.shape

    padded_A = np.empty(shape=(rows + size - 1, cols + size - 1),
                        dtype=M.dtype)
    padded_A[:] = np.nan
    rows_pad, cols_pad = padded_A.shape

    if no_data_val:
        mask = M == no_data_val
        M[mask] = np.nan

    upleft_data_start = int((size - 1) / 2)
    rl, cl = rows_pad - upleft_data_start, cols_pad - upleft_data_start
    padded_A[upleft_data_start:rl, upleft_data_start:cl] = M.copy()

    n, m = M.shape
    strided_data = as_strided(padded_A, (n, m, size, size),
                              padded_A.strides + padded_A.strides)
    strided_data = strided_data.copy().reshape((n, m, size**2))

    return func(strided_data, axis=2)


def find_masked_rows(m, masking_value=0):
    """Find rows in m filled with masking value.

    :param m: Numpy matrix to be searched for masked rows.
    :param masking_value: Value that corresponds to missing contacts,
                          defaults to 0
    :returns: Indices of masked rows.
    """
    s = np.sum(m, 0)
    n_bins = m.shape[0]
    cutoff = n_bins * masking_value
    idxs = np.where(s == cutoff)[0]

    return idxs


def remove_rows(m, target_rows):
    """Remove target bins from m.

    Remove rows and columns with indices indicated in target rows form m.
    :param m: Numpy matrix.
    :param target_rows: List of indices of bins to be removed from m.
    :returns: Truncated numpy matrix.
    """
    idxs = target_rows
    m_removed = np.delete(m, idxs, 0)
    m_removed = np.delete(m_removed, idxs, 1)

    return m_removed


def comparison_worker(input_queue,
                      output_queue,
                      reference_edges,
                      reference_regions,
                      query_edges,
                      query_regions,
                      min_bins=20,
                      keep_unmappable_bins=False,
                      absolute_window_size=None,
                      relative_window_size=1.,
                      mappability_cutoff=0.1,
                      default_value=1):
    worker_id = uuid.uuid4()
    try:
        while True:
            data = input_queue.get(block=True)
            if data is None:
                break
            pair, compute_sn = data

            pair_ix, ssim, sn = compare_pair(
                pair,
                reference_edges,
                reference_regions,
                query_edges,
                query_regions,
                min_bins=min_bins,
                keep_unmappable_bins=keep_unmappable_bins,
                absolute_window_size=absolute_window_size,
                relative_window_size=relative_window_size,
                mappability_cutoff=mappability_cutoff,
                default_value=default_value,
                compute_sn=compute_sn)
            output_queue.put((pair_ix, ssim, sn))
    except Exception as e:
        logger.error("Worker {} encountered a problem: {}".format(
            worker_id, e))
        output_queue.put(e)

    logger.info("Worker {} exited gracefully".format(worker_id))


def chunk_comparison_worker(input_queue,
                            output_queue,
                            reference_edges,
                            reference_regions,
                            query_edges,
                            query_regions,
                            min_bins=20,
                            keep_unmappable_bins=False,
                            absolute_window_size=None,
                            relative_window_size=1.,
                            mappability_cutoff=0.1,
                            default_value=1):
    worker_id = uuid.uuid4()
    previous_reference = [None, (None, None)]
    previous_query = [None, (None, None)]
    try:
        while True:
            data = input_queue.get(block=True)
            if data is None:
                break
            chunk, compute_sn = data

            results = []
            for pair in chunk:
                pair_ix, reference_region, query_region = pair
                try:
                    if previous_reference[
                            0] is not None and reference_region == previous_reference[
                                0]:
                        reference, ref_rs = previous_reference[1]
                    else:
                        reference, ref_rs = sub_matrix_from_edges_dict(
                            reference_edges,
                            reference_regions,
                            reference_region,
                            default_weight=default_value)
                        previous_reference = [
                            reference_region, (reference, ref_rs)
                        ]

                    if previous_query[
                            0] is not None and query_region == previous_query[
                                0]:
                        query, qry_rs = previous_query[1]
                    else:
                        query, qry_rs = sub_matrix_from_edges_dict(
                            query_edges,
                            query_regions,
                            query_region,
                            default_weight=default_value)
                        previous_query = [query_region, (query, qry_rs)]
                except ValueError:
                    results.append([pair_ix, np.nan, np.nan])
                    continue

                pair_ix, ssim, sn = compare_matrix_pair(
                    reference,
                    reference_region,
                    query,
                    query_region,
                    pair_ix=pair_ix,
                    min_bins=min_bins,
                    keep_unmappable_bins=keep_unmappable_bins,
                    absolute_window_size=absolute_window_size,
                    relative_window_size=relative_window_size,
                    mappability_cutoff=mappability_cutoff,
                    default_value=default_value,
                    compute_sn=compute_sn)
                results.append([pair_ix, ssim, sn])
            output_queue.put(results)
    except Exception as e:
        logger.error("Worker {} encountered a problem: {}".format(
            worker_id, e))
        output_queue.put(e)

    logger.debug("Worker {} exited gracefully".format(worker_id))


def compare_pair(pair,
                 reference_edges,
                 reference_regions,
                 query_edges,
                 query_regions,
                 min_bins=20,
                 keep_unmappable_bins=False,
                 absolute_window_size=None,
                 relative_window_size=1.,
                 mappability_cutoff=0.1,
                 default_value=1,
                 compute_sn=True):
    """Run comparison of given Hi-C matrices without any background model.

    Compare given submatrices of the full Hi-C matrices in sampleID2hic.
    All submatrices (specified in pairs) have to be of the same size.
    The assignment of reference and query samples is arbitrary in this case,
    but has to be consistent.
    :param edges: edges dictionary
    :param regions: regions intervaltree

    :param worker_ID: String or int used as ID for the process in which this
                      function is run.
    :param pairs: Dictionary indicating the submatrix pairs that
                  will be compared.
                  Must have form:
                  {
                    pair_id: (reference_region, query_region)
                  },
                  where reference_region and query_region
                  are :class:`~GenomicRegion` objects.
    :param min_bins: Minimum size of matrix to be considered for comparison,
                     defaults to 20
    :param keep_unmappable_bins: If True, disables removal of unmappable bins
                                 from matrices prior to comparison,
                                 defaults to False
    :param absolute_windowsize: Absolute window size used in the
                                structural similarity function. Overwrites
                                relative_windowsize, defaults to None
    :param relative_windowsize: Window size relative to the size
                                of the compared matrices,
                                used in the structural similarity function,
                                defaults to 1.
    :param mappability_cutoff: Maximum fraction of unmappable bins allowed
                               in a matrix to be considered for comparison,
                               defaults to 0.1

    :returns: results dictionary of form:
              {
                pair_id: (similarity_score, signal_to_noise_value)
              }
    """
    pair_ix, reference_region, query_region = pair
    try:
        reference, ref_rs = sub_matrix_from_edges_dict(
            reference_edges,
            reference_regions,
            reference_region,
            default_weight=default_value)
        query, qry_rs = sub_matrix_from_edges_dict(
            query_edges,
            query_regions,
            query_region,
            default_weight=default_value)
    except ValueError:
        return pair_ix, np.nan, np.nan

    return compare_matrix_pair(reference,
                               reference_region,
                               query,
                               query_region,
                               pair_ix=pair_ix,
                               min_bins=min_bins,
                               keep_unmappable_bins=keep_unmappable_bins,
                               absolute_window_size=absolute_window_size,
                               relative_window_size=relative_window_size,
                               mappability_cutoff=mappability_cutoff,
                               default_value=default_value,
                               compute_sn=compute_sn)


def compare_matrix_pair(reference,
                        reference_region,
                        query,
                        query_region,
                        pair_ix=None,
                        min_bins=20,
                        keep_unmappable_bins=False,
                        absolute_window_size=None,
                        relative_window_size=1.,
                        mappability_cutoff=0.1,
                        default_value=1,
                        compute_sn=True):
    # check size criteria
    if query.shape[0] < min_bins or reference.shape[0] < min_bins:
        return pair_ix, np.nan, np.nan

    # resize matrices if needed
    size_factor = len(reference) / len(query)
    if size_factor < 1:
        reference = resize(reference, np.shape(query), order=0,
                           mode='reflect').astype(np.float64)
    elif size_factor > 1:
        query = resize(query, np.shape(reference), order=0,
                       mode='reflect').astype(np.float64)

    # check if either matrix needs to be flipped
    if reference_region.strand != query_region.strand:
        query = np.fliplr(np.flipud(query))

    # check content in unmappable bins in input matrices
    idxs = []
    init_filter_pass = True
    if mappability_cutoff > 0:
        ms = [reference, query]
        for m in ms:
            m = np.array(m)
            idx = find_masked_rows(m, masking_value=default_value)
            idxs.append(idx)
            if len(idx) / np.shape(m)[0] > mappability_cutoff:
                init_filter_pass = False
    if not init_filter_pass:
        return pair_ix, np.nan, np.nan

    # remove masked bins
    if not keep_unmappable_bins:
        target_rows = np.union1d(*idxs)
        reference = remove_rows(reference, target_rows)
        query = remove_rows(query, target_rows)

    # compare!
    if absolute_window_size:
        window_size = absolute_window_size
    else:
        window_size = int(np.shape(reference)[0] * relative_window_size)
    if window_size % 2 == 0:
        window_size -= 1

    # prevent from going lower than three.
    # breaks at ws = 1 because of 0 division, and 2 is not odd.
    window_size = max(window_size, 3)

    # actual comparison
    dmax = max(reference.max(), query.max())
    dmin = min(reference.min(), query.min())
    drange = dmax - dmin
    curr_ssim = structural_similarity(reference,
                                      query,
                                      win_size=window_size,
                                      data_range=drange)
    if compute_sn:
        curr_sn = SN(reference, query)
    else:
        curr_sn = None

    return pair_ix, curr_ssim, curr_sn


def compare_structures(reference_edges,
                       reference_regions,
                       query_edges,
                       query_regions,
                       pairs,
                       worker_ID,
                       min_bins=20,
                       keep_unmappable_bins=False,
                       absolute_window_size=None,
                       relative_window_size=1.,
                       mappability_cutoff=0.1,
                       default_value=1):
    """Run comparison of given Hi-C matrices without any background model.

    Compare given submatrices of the full Hi-C matrices in sampleID2hic.
    All submatrices (specified in pairs) have to be of the same size.
    The assignment of reference and query samples is arbitrary in this case,
    but has to be consistent.
    :param edges: edges dictionary
    :param regions: regions intervaltree

    :param worker_ID: String or int used as ID for the process in which this
                      function is run.
    :param pairs: Dictionary indicating the submatrix pairs that
                  will be compared.
                  Must have form:
                  {
                    pair_id: (reference_region, query_region)
                  },
                  where reference_region and query_region
                  are :class:`~GenomicRegion` objects.
    :param min_bins: Minimum size of matrix to be considered for comparison,
                     defaults to 20
    :param keep_unmappable_bins: If True, disables removal of unmappable bins
                                 from matrices prior to comparison,
                                 defaults to False
    :param absolute_windowsize: Absolute window size used in the
                                structural similarity function. Overwrites
                                relative_windowsize, defaults to None
    :param relative_windowsize: Window size relative to the size
                                of the compared matrices,
                                used in the structural similarity function,
                                defaults to 1.
    :param mappability_cutoff: Maximum fraction of unmappable bins allowed
                               in a matrix to be considered for comparison,
                               defaults to 0.1
    :returns: results dictionary of form:
              {
                pair_id: (similarity_score, signal_to_noise_value)
              }
    """

    results = {}  # store the ssim in here
    no_pass = set()

    for pair in pairs:
        pair_ix, ssim, sn = compare_pair(
            pair,
            reference_edges,
            reference_regions,
            query_edges,
            query_regions,
            min_bins=min_bins,
            keep_unmappable_bins=keep_unmappable_bins,
            absolute_window_size=absolute_window_size,
            relative_window_size=relative_window_size,
            mappability_cutoff=mappability_cutoff,
            default_value=default_value)
        if np.isnan(ssim):
            no_pass.add(pair_ix)
            results[pair_ix] = (np.nan, np.nan)
        else:
            results[pair_ix] = (ssim, sn)

    logger.debug('[WORKER #{}]: Done!'.format(worker_ID))

    return results



import os
from setuptools import setup, find_packages, Command


__version__ = None
exec(open('chess/version.py').read())


class CleanCommand(Command):
    """
    Custom clean command to tidy up the project root.
    """
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info ./htmlcov')


def read_file(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as file:
        return file.read()


setup(
    name='chess-hic',
    version=__version__,
    description='Quantitative comparison and automatic feature extraction for chromatin contact data.',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/vaquerizaslab/chess',
    setup_requires=[
        'setuptools>=18.0',
    ],
    packages=find_packages(),
    install_requires=[
        'cython',
        'numpy>=1.8.0',
        'fanc>=0.9.9',
        'scikit-image',
        'future',
        'pathos',
        'pandas>=1.1.3',
        'scipy',
        'intervaltree',
        'pybedtools',
        'tqdm',
        'kneed',
    ],
    scripts=['bin/chess'],
    cmdclass={
        'clean': CleanCommand
    },
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)
